<?php

namespace Phpactor\Tests\Unit\Extension\ContextMenu\Handler;

use Phpactor\CodeTransform\Domain\Helper\InterestingOffsetFinder;
use Phpactor\Extension\ContextMenu\Model\ContextMenu;
use Phpactor\Extension\Rpc\Handler;
use Phpactor\Extension\ContextMenu\Handler\ContextMenuHandler;
use Phpactor\TextDocument\ByteOffset;
use Phpactor\WorseReflection\Reflector;
use Phpactor\WorseReflection\Core\SourceCode;
use Phpactor\Extension\Rpc\Response\EchoResponse;
use Phpactor\WorseReflection\Core\Inference\Symbol;
use Phpactor\Extension\Rpc\Response\InputCallbackResponse;
use Phpactor\Extension\Rpc\Request;
use Phpactor\Extension\ContextMenu\ContextMenuExtension;
use Phpactor\Extension\Rpc\RequestHandler\RequestHandler;
use Phpactor\Extension\Core\Application\Helper\ClassFileNormalizer;
use Phpactor\WorseReflection\ReflectorBuilder;
use Phpactor\Container\Container;
use Phpactor\Tests\Unit\Extension\Rpc\HandlerTestCase;

class ContextMenuHandlerTest extends HandlerTestCase
{
    const VARIABLE_ACTION = 'do_something';
    const SOURCE = '<?php $hello = "world"; echo $hello;';
    const FOUND_OFFSET = 10;
    const ORIGINAL_OFFSET = 8;

    /**
     * @var Reflector
     */
    private $reflector;

    /**
     * @var Container
     */
    private $container;

    /**
     * @var array
     */
    private $menu = [];

    /**
     * @var RequestHandler
     */
    private $requestHandler;

    /**
     * @var ClassFileNormalizer
     */
    private $classFileNormalizer;

    /**
     * @var InterestingOffsetFinder
     */
    private $offsetFinder;

    public function setUp(): void
    {
        $this->reflector = ReflectorBuilder::create()->addSource(SourceCode::fromPath(__FILE__))->build();
        $this->offsetFinder = $this->prophesize(InterestingOffsetFinder::class);
        $this->classFileNormalizer = $this->prophesize(ClassFileNormalizer::class);
        $this->container = $this->prophesize(Container::class);
        $this->requestHandler = $this->prophesize(RequestHandler::class);
    }

    public function createHandler(): Handler
    {
        return new ContextMenuHandler(
            $this->reflector,
            $this->offsetFinder->reveal(),
            $this->classFileNormalizer->reveal(),
            $this->menu,
            $this->container->reveal()
        );
    }

    public function testNoActionsAvailable(): void
    {
        $this->menu = ContextMenu::fromArray([
            'actions' => [
                Symbol::VARIABLE => [
                    'action' => self::VARIABLE_ACTION,
                    'parameters' => [
                        'one' => 1,
                    ],
                ]
            ],
            'contexts' => [
                Symbol::VARIABLE => [
                ]
            ]
        ]);
        $source = SourceCode::fromPathAndString(
            '/hello.php',
            '<?php $hello = "world"; echo $hello;'
        );
        $offset = ByteOffset::fromInt(4);

        $this->offsetFinder->find($source, $offset)
            ->willReturn($offset);

        $action = $this->handle(ContextMenuHandler::NAME, [
            'source' => (string) $source,
            'offset' => $offset->toInt(),
            'current_path' => $source->path(),
        ]);

        $this->assertInstanceOf(EchoResponse::class, $action);
        $this->assertStringContainsString('No context actions', $action->message());
    }

    public function testReturnMenu(): void
    {
        $this->menu = ContextMenu::fromArray([
            'actions' => [
                Symbol::VARIABLE => [
                    'action' => self::VARIABLE_ACTION,
                    'parameters' => [
                        'one' => 1,
                    ],
                ]
            ],
            'contexts' => [
                Symbol::VARIABLE => [
                    Symbol::VARIABLE
                ]
            ]
        ]);

        $source = SourceCode::fromPathAndString(
            '/hello.php',
            '<?php $hello = "world"; echo $hello;'
        );
        $offset = ByteOffset::fromInt(self::ORIGINAL_OFFSET);

        $this->offsetFinder->find($source, $offset)
            ->willReturn($offset);

        $action = $this->handle(ContextMenuHandler::NAME, [
            'source' => (string) $source,
            'offset' => $offset->toInt(),
            'current_path' => $source->path(),
        ]);

        $this->assertInstanceOf(InputCallbackResponse::class, $action);
        $this->assertInstanceOf(Request::class, $action->callbackAction());
        $this->assertEquals(ContextMenuHandler::NAME, $action->callbackAction()->name());
    }

    public function testReturnMenuWithOriginalOffset(): void
    {
        $this->menu = ContextMenu::fromArray([
            'actions' => [
                Symbol::VARIABLE => [
                    'action' => self::VARIABLE_ACTION,
                    'parameters' => [
                        'one' => 1,
                    ],
                ]
            ],
            'contexts' => [
                Symbol::VARIABLE => [
                    Symbol::VARIABLE
                ]
            ]
        ]);

        $source = SourceCode::fromPathAndString(
            '/hello.php',
            '<?php $hello = "world"; echo $hello;'
        );
        $offset = ByteOffset::fromInt(self::ORIGINAL_OFFSET);

        $this->offsetFinder->find($source, $offset)
            ->willReturn(ByteOffset::fromInt(self::FOUND_OFFSET));

        $action = $this->handle(ContextMenuHandler::NAME, [
            'source' => (string) $source,
            'offset' => self::ORIGINAL_OFFSET,
            'current_path' => $source->path(),
        ]);

        $this->assertInstanceOf(InputCallbackResponse::class, $action);
        $this->assertEquals(self::ORIGINAL_OFFSET, $action->callbackAction()->parameters()['offset']);
    }

    public function testReplaceTokens(): void
    {
        $this->container->get(ContextMenuExtension::SERVICE_REQUEST_HANDLER)->willReturn(
            $this->requestHandler->reveal()
        );

        $this->classFileNormalizer->classToFile('string')->willReturn(__FILE__);

        $source = SourceCode::fromPathAndString('/hello.php', self::SOURCE);
        $offset = ByteOffset::fromInt(self::ORIGINAL_OFFSET);

        $this->offsetFinder->find($source, $offset)
            ->willReturn($offset);

        $this->requestHandler->handle(
            Request::fromNameAndParameters(
                self::VARIABLE_ACTION,
                [
                    'some_source' => (string) $source,
                    'some_offset' => $offset->toInt(),
                    'some_path' => __FILE__
                ]
            )
        )->willReturn(
            EchoResponse::fromMessage('Hello')
        );

        $this->menu = ContextMenu::fromArray([
            'actions' => [
                self::VARIABLE_ACTION => [
                    'action' => self::VARIABLE_ACTION,
                    'parameters' => [
                        'some_source' => '%source%',
                        'some_offset' => '%offset%',
                        'some_path' => '%path%',
                    ],
                ]
            ],
            'contexts' => [
                Symbol::VARIABLE => [
                    self::VARIABLE_ACTION
                ]
            ]
        ]);

        $action = $this->handle(ContextMenuHandler::NAME, [
            'action' => self::VARIABLE_ACTION,
            'source' => (string) $source,
            'offset' => $offset->toInt(),
            'current_path' => $source->path(),
        ]);

        $parameters = $action->parameters();
        $this->assertEquals([
            'message' => 'Hello',
        ], $parameters);
    }
}
